@@ -29,6 +29,7 @@
 using System;
 using System.Collections;
 using System.Runtime.InteropServices;
+using System.Threading;
 using Mono.Unix;
 using Hyena;
 using Hyena.Data;
@@ -57,6 +58,8 @@ namespace Banshee.GStreamer
     internal delegate void BansheePlayerIterateCallback (IntPtr player);
     internal delegate void BansheePlayerBufferingCallback (IntPtr player, int buffering_progress);
     internal delegate void BansheePlayerVisDataCallback (IntPtr player, int channels, int samples, IntPtr data, int bands, IntPtr spectrum);
+    internal delegate void BansheePlayerNextTrackStartingCallback (IntPtr player);
+    internal delegate void BansheePlayerAboutToFinishCallback (IntPtr player);
     internal delegate IntPtr VideoPipelineSetupHandler (IntPtr player, IntPtr bus);
 
     internal delegate void GstTaggerTagFoundCallback (IntPtr player, string tagName, ref GLib.Value value);
@@ -73,6 +76,7 @@ namespace Banshee.GStreamer
         private uint GST_STREAM_ERROR = 0;
 
         private HandleRef handle;
+        private bool initialized;
 
         private BansheePlayerEosCallback eos_callback;
         private BansheePlayerErrorCallback error_callback;
@@ -82,11 +86,19 @@ namespace Banshee.GStreamer
         private BansheePlayerVisDataCallback vis_data_callback;
         private VideoPipelineSetupHandler video_pipeline_setup_callback;
         private GstTaggerTagFoundCallback tag_found_callback;
+        private BansheePlayerNextTrackStartingCallback next_track_starting_callback;
+        private BansheePlayerAboutToFinishCallback about_to_finish_callback;
+
+        private bool next_track_pending;
+        private SafeUri pending_uri;
 
         private bool buffering_finished;
         private int pending_volume = -1;
         private bool xid_is_set = false;
 
+        private bool gapless_enabled;
+        private EventWaitHandle next_track_set;
+
         private event VisualizationDataHandler data_available = null;
         public event VisualizationDataHandler DataAvailable {
             add {
@@ -138,7 +150,8 @@ namespace Banshee.GStreamer
             vis_data_callback = new BansheePlayerVisDataCallback (OnVisualizationData);
             video_pipeline_setup_callback = new VideoPipelineSetupHandler (OnVideoPipelineSetup);
             tag_found_callback = new GstTaggerTagFoundCallback (OnTagFound);
-
+            next_track_starting_callback = new BansheePlayerNextTrackStartingCallback (OnNextTrackStarting);
+            about_to_finish_callback = new BansheePlayerAboutToFinishCallback (OnAboutToFinish);
             bp_set_eos_callback (handle, eos_callback);
 #if !WIN32
             bp_set_iterate_callback (handle, iterate_callback);
@@ -147,7 +160,10 @@ namespace Banshee.GStreamer
             bp_set_state_changed_callback (handle, state_changed_callback);
             bp_set_buffering_callback (handle, buffering_callback);
             bp_set_tag_found_callback (handle, tag_found_callback);
+            bp_set_next_track_starting_callback (handle, next_track_starting_callback);
             bp_set_video_pipeline_setup_callback (handle, video_pipeline_setup_callback);
+
+            next_track_set = new EventWaitHandle (false, EventResetMode.ManualReset);
         }
 
         protected override void Initialize ()
@@ -158,6 +174,7 @@ namespace Banshee.GStreamer
                 throw new ApplicationException (Catalog.GetString ("Could not initialize GStreamer library"));
             }
 
+            initialized = true;
             OnStateChanged (PlayerState.Ready);
 
             if (pending_volume >= 0) {
@@ -166,6 +183,7 @@ namespace Banshee.GStreamer
 
             InstallPreferences ();
             ReplayGainEnabled = ReplayGainEnabledSchema.Get ();
+            GaplessEnabled = GaplessEnabledSchema.Get ();
         }
 
         public override void Dispose ()
@@ -215,6 +233,30 @@ namespace Banshee.GStreamer
             bp_pause (handle);
         }
 
+        public override void SetNextTrackUri (SafeUri uri)
+        {
+            next_track_pending = false;
+            if (next_track_set.WaitOne (0)) {
+                // We're not waiting for the next track to be set.
+                // This means that we've missed the window for gapless.
+                // Save this URI to be played when we receive EOS.
+                pending_uri = uri;
+                return;
+            }
+            // If there isn't a next track for us, release the block on the about-to-finish callback.
+            if (uri == null) {
+                next_track_set.Set ();
+                return;
+            }
+            IntPtr uri_ptr = GLib.Marshaller.StringToPtrGStrdup (uri.AbsoluteUri);
+            try {
+                bp_set_next_track (handle, uri_ptr);
+            } finally {
+                GLib.Marshaller.Free (uri_ptr);
+                next_track_set.Set ();
+            }
+        }
+
         public override void VideoExpose (IntPtr window, bool direct)
         {
             bp_video_window_expose (handle, window, direct);
@@ -240,6 +282,57 @@ namespace Banshee.GStreamer
         {
             Close (false);
             OnEventChanged (PlayerEvent.EndOfStream);
+            if (!next_track_pending) {
+                OnEventChanged (PlayerEvent.RequestNextTrack);
+            } else if (pending_uri != null) {
+                Log.Warning ("[Gapless] EOS signalled while waiting for next track.  This means that Banshee " +
+                    "was too slow at calculating what track to play next.  " +
+                    "If this happens frequently, please file a bug");
+                OnStateChanged (PlayerState.Loading);
+                OpenUri (pending_uri);
+                Play ();
+                pending_uri = null;
+            } else {
+                // This should be unreachable - the RequestNextTrack event is delegated to the main thread
+                // and so blocks the bus callback from delivering the EOS message.
+                //
+                // Playback should continue as normal from here, when the RequestNextTrack message gets handled.
+                Log.Warning ("[Gapless] EndOfStream message received before the next track has been set.  " +
+                    "If this happens frequently, please file a bug");
+            }
+        }
+
+        private void OnNextTrackStarting (IntPtr player)
+        {
+            if (GaplessEnabled) {
+                OnEventChanged (PlayerEvent.EndOfStream);
+                OnEventChanged (PlayerEvent.StartOfStream);
+            }
+        }
+
+        private void OnAboutToFinish (IntPtr player)
+        {
+            // This is needed to make Shuffle-by-* work.
+            // Shuffle-by-* uses the LastPlayed field to determine what track in the grouping to play next.
+            // Therefore, we need to update this before requesting the next track.
+            //
+            // This will be overridden by IncrementLastPlayed () called by
+            // PlaybackControllerService's EndOfStream handler.
+            CurrentTrack.LastPlayed = DateTime.Now;
+            CurrentTrack.Save ();
+
+            next_track_set.Reset ();
+            pending_uri = null;
+            next_track_pending = true;
+
+            OnEventChanged (PlayerEvent.RequestNextTrack);
+            // Gapless playback with Playbin2 requires that the about-to-finish callback does not return until
+            // the next uri has been set.  Block here for a second or until the RequestNextTrack event has
+            // finished triggering.
+            if (!next_track_set.WaitOne (1000, false)) {
+                Log.Debug ("[Gapless] Timed out while waiting for next_track_set to be raised");
+                next_track_set.Set ();
+            }
         }
 
         private void OnIterate (IntPtr player)
@@ -249,7 +342,11 @@ namespace Banshee.GStreamer
 
         private void OnStateChange (IntPtr player, GstState old_state, GstState new_state, GstState pending_state)
         {
-            if (old_state == GstState.Ready && new_state == GstState.Paused && pending_state == GstState.Playing) {
+            if (CurrentState != PlayerState.Loaded && old_state == GstState.Ready && new_state == GstState.Paused && pending_state == GstState.Playing) {
+                if (ready_timeout != 0) {
+                    Application.IdleTimeoutRemove (ready_timeout);
+                    ready_timeout = 0;
+                }
                 OnStateChanged (PlayerState.Loaded);
                 return;
             } else if (old_state == GstState.Paused && new_state == GstState.Playing && pending_state == GstState.VoidPending) {
@@ -261,9 +358,30 @@ namespace Banshee.GStreamer
             } else if (CurrentState == PlayerState.Playing && old_state == GstState.Playing && new_state == GstState.Paused) {
                 OnStateChanged (PlayerState.Paused);
                 return;
+            } else if (new_state == GstState.Ready && pending_state == GstState.Playing) {
+                if (ready_timeout == 0) {
+                    ready_timeout = Application.RunTimeout (1000, OnReadyTimeout);
+                }
+                return;
             }
         }
 
+        private uint ready_timeout;
+        private bool OnReadyTimeout ()
+        {
+            ready_timeout = 0;
+            var uri = CurrentUri;
+            if (CurrentState == PlayerState.Loading && uri != null && uri.IsLocalPath) {
+                // This is a dirty workaround.  I was seeing the playback get stuck on track transition,
+                // about one in 20 songs, where it would load the new track's duration, but be stuck at 0:00,
+                // but if I moved the seek slider it would unstick it, hence setting Position...
+                Log.WarningFormat ("Seem to be stuck loading {0}, so re-trying", uri);
+                Position = 0;
+            }
+
+            return false;
+        }
+
         private void OnError (IntPtr player, uint domain, int code, IntPtr error, IntPtr debug)
         {
             Close (true);
@@ -391,7 +509,7 @@ namespace Banshee.GStreamer
         public override ushort Volume {
             get { return (ushort)Math.Round (bp_get_volume (handle) * 100.0); }
             set {
-                if ((IntPtr)handle == IntPtr.Zero) {
+                if (!initialized) {
                     pending_volume = value;
                     return;
                 }
@@ -499,6 +617,24 @@ namespace Banshee.GStreamer
             set { bp_replaygain_set_enabled (handle, value); }
         }
 
+        private bool GaplessEnabled {
+            get { return gapless_enabled; }
+            set
+            {
+                if (bp_supports_gapless (handle)) {
+                    gapless_enabled = value;
+                    if (value) {
+                        bp_set_about_to_finish_callback (handle, about_to_finish_callback);
+                    } else {
+                        bp_set_about_to_finish_callback (handle, null);
+                    }
+                } else {
+                    gapless_enabled = false;
+                }
+            }
+        }
+
+
 #region ISupportClutter
 
         private IntPtr clutter_video_sink;
@@ -525,6 +661,7 @@ namespace Banshee.GStreamer
             }
         }
 
+
         private IntPtr OnVideoPipelineSetup (IntPtr player, IntPtr bus)
         {
             try {
@@ -553,6 +690,7 @@ namespace Banshee.GStreamer
 #region Preferences
 
         private PreferenceBase replaygain_preference;
+        private PreferenceBase gapless_preference;
 
         private void InstallPreferences ()
         {
@@ -566,6 +704,13 @@ namespace Banshee.GStreamer
                 Catalog.GetString ("For tracks that have ReplayGain data, automatically scale (normalize) playback volume"),
                 delegate { ReplayGainEnabled = ReplayGainEnabledSchema.Get (); }
             ));
+            if (bp_supports_gapless (handle)) {
+                gapless_preference = service["general"]["misc"].Add (new SchemaPreference<bool> (GaplessEnabledSchema,
+                        Catalog.GetString ("Enable _gapless playback"),
+                        Catalog.GetString ("Eliminate the small playback gap on track change.  Useful for concept albums and classical music."),
+                        delegate { GaplessEnabled = GaplessEnabledSchema.Get (); }
+                ));
+            }
         }
 
         private void UninstallPreferences ()
@@ -576,7 +721,11 @@ namespace Banshee.GStreamer
             }
 
             service["general"]["misc"].Remove (replaygain_preference);
+            if (bp_supports_gapless (handle)) {
+                service["general"]["misc"].Remove (gapless_preference);
+            }
             replaygain_preference = null;
+            gapless_preference = null;
         }
 
         public static readonly SchemaEntry<bool> ReplayGainEnabledSchema = new SchemaEntry<bool> (
@@ -586,6 +735,14 @@ namespace Banshee.GStreamer
             "If ReplayGain data is present on tracks when playing, allow volume scaling"
         );
 
+        public static readonly SchemaEntry<bool> GaplessEnabledSchema = new SchemaEntry<bool> (
+            "player_engine", "gapless_playback_enabled",
+            true,
+            "Enable gapless playback",
+            "Eliminate the small playback gap on track change.  Useful for concept albums & classical music."
+        );
+
+
 #endregion
 
         [DllImport ("libbanshee.dll")]
@@ -626,6 +783,17 @@ namespace Banshee.GStreamer
         private static extern void bp_set_tag_found_callback (HandleRef player,
             GstTaggerTagFoundCallback cb);
 
+        [DllImport ("libbanshee.dll")]
+        private static extern void bp_set_next_track_starting_callback (HandleRef player,
+            BansheePlayerNextTrackStartingCallback cb);
+
+        [DllImport ("libbanshee.dll")]
+        private static extern void bp_set_about_to_finish_callback (HandleRef player,
+            BansheePlayerAboutToFinishCallback cb);
+
+        [DllImport ("libbanshee.dll")]
+        private static extern bool bp_supports_gapless (HandleRef player);
+
         [DllImport ("libbanshee.dll")]
         private static extern bool bp_open (HandleRef player, IntPtr uri);
 
@@ -638,6 +806,9 @@ namespace Banshee.GStreamer
         [DllImport ("libbanshee.dll")]
         private static extern void bp_play (HandleRef player);
 
+        [DllImport ("libbanshee.dll")]
+        private static extern bool bp_set_next_track (HandleRef player, IntPtr uri);
+
         [DllImport ("libbanshee.dll")]
         private static extern void bp_set_volume (HandleRef player, double volume);
 
