@@ -66,14 +66,19 @@ namespace Banshee.Streaming
 
         private static string Choose (string priority, string fallback, bool flip)
         {
-            return flip
+            return TrimZero (flip
                 ? IsNullOrEmpty (fallback) ? priority : fallback
-                : IsNullOrEmpty (priority) ? fallback : priority;
+                : IsNullOrEmpty (priority) ? fallback : priority);
         }
 
         private static bool IsNullOrEmpty (string s)
         {
-            return String.IsNullOrEmpty (s) || s.IndexOf ('\0') >= 0 && s.Trim ('\0').Length == 0;
+            return String.IsNullOrEmpty (s) || TrimZero (s).Length == 0;
+        }
+
+        private static string TrimZero (string s)
+        {
+            return s == null || s.IndexOf ('\0') == -1 ? s : s.Split ('\0')[0];
         }
 
         #pragma warning disable 0169
@@ -149,6 +154,8 @@ namespace Banshee.Streaming
                 track.MimeType = file.MimeType;
                 track.Duration = file.Properties.Duration;
                 track.BitRate  = file.Properties.AudioBitrate;
+                track.SampleRate  = file.Properties.AudioSampleRate;
+                track.BitsPerSample  = file.Properties.BitsPerSample;
 
                 FindTrackMediaAttributes (track, file);
 
@@ -158,7 +165,7 @@ namespace Banshee.Streaming
                 track.AlbumTitleSort = Choose (file.Tag.AlbumSort, track.AlbumTitleSort, preferTrackInfo);
                 track.AlbumArtist = Choose (file.Tag.FirstAlbumArtist, track.AlbumArtist, preferTrackInfo);
                 track.AlbumArtistSort = Choose (file.Tag.FirstAlbumArtistSort, track.AlbumArtistSort, preferTrackInfo);
-                track.IsCompilation = IsCompilation (file);
+                track.IsCompilation = preferTrackInfo ? track.IsCompilation : IsCompilation (file);
 
                 track.TrackTitle = Choose (file.Tag.Title, track.TrackTitle, preferTrackInfo);
                 track.TrackTitleSort = Choose (file.Tag.TitleSort, track.TrackTitleSort, preferTrackInfo);
@@ -184,7 +191,7 @@ namespace Banshee.Streaming
                 }
             } else {
                 track.MediaAttributes = TrackMediaAttributes.AudioStream;
-                if (track.Uri != null && VideoExtensions.IsMatchingFile (track.Uri.LocalPath)) {
+                if (track.Uri != null && VideoExtensions.IsMatchingFile (track.Uri.AbsoluteUri)) {
                     track.MediaAttributes = TrackMediaAttributes.VideoStream;
                 }
             }
@@ -195,7 +202,7 @@ namespace Banshee.Streaming
 
             if (String.IsNullOrEmpty (track.TrackTitle)) {
                 try {
-                    string filename = System.IO.Path.GetFileNameWithoutExtension (track.Uri.LocalPath);
+                    string filename = System.IO.Path.GetFileNameWithoutExtension (track.Uri.AbsoluteUri);
                     if (!String.IsNullOrEmpty (filename)) {
                         track.TrackTitle = filename;
                     }
@@ -209,6 +216,12 @@ namespace Banshee.Streaming
 
         private static bool IsCompilation (TagLib.File file)
         {
+            try {
+                var xiph_tag = file.GetTag(TagLib.TagTypes.Xiph, true) as TagLib.Ogg.XiphComment;
+                if (xiph_tag != null && xiph_tag.IsCompilation)
+                    return true;
+            } catch {}
+
             try {
                 TagLib.Id3v2.Tag id3v2_tag = file.GetTag(TagLib.TagTypes.Id3v2, true) as TagLib.Id3v2.Tag;
                 if (id3v2_tag != null && id3v2_tag.IsCompilation)
@@ -234,6 +247,14 @@ namespace Banshee.Streaming
 
         private static void SaveIsCompilation (TagLib.File file, bool is_compilation)
         {
+            try {
+                var xiph_tag = file.GetTag(TagLib.TagTypes.Xiph, true) as TagLib.Ogg.XiphComment;
+                if (xiph_tag != null) {
+                    xiph_tag.IsCompilation = is_compilation;
+                    return;
+                }
+            } catch {}
+
             try {
                 TagLib.Id3v2.Tag id3v2_tag = file.GetTag(TagLib.TagTypes.Id3v2, true) as TagLib.Id3v2.Tag;
                 if (id3v2_tag != null) {
