@@ -34,6 +34,7 @@ using System.Linq;
 
 using Mono.Unix;
 
+using Hyena;
 using Hyena.Collections;
 using Hyena.Data.Sqlite;
 
@@ -58,7 +59,8 @@ namespace Banshee.PlayQueue
 
         private ITrackModelSource prior_playback_source;
         private DatabaseTrackInfo current_track;
-        private long offset;
+        private Shuffler shuffler;
+        private long offset = -1;
         private TrackInfo prior_playback_track;
         private PlayQueueActions actions;
         private bool was_playing = false;
@@ -68,19 +70,19 @@ namespace Banshee.PlayQueue
         private SourcePage pref_page;
         private Section pref_section;
 
-        private PlaybackShuffleMode populate_mode = (PlaybackShuffleMode) PopulateModeSchema.Get ();
+        private string populate_shuffle_mode = PopulateModeSchema.Get ();
         private string populate_from_name = PopulateFromSchema.Get ();
-        private ITrackModelSource populate_from = null;
+        private DatabaseSource populate_from = null;
         private int played_songs_number = PlayedSongsNumberSchema.Get ();
         private int upcoming_songs_number = UpcomingSongsNumberSchema.Get ();
-        
+
         public PlayQueueSource () : base (Catalog.GetString ("Play Queue"), null)
         {
             BindToDatabase ();
             TypeUniqueId = DbId.ToString ();
             Initialize ();
             AfterInitialized ();
-            
+
             Order = 20;
             Properties.SetString ("Icon.Name", "source-playlist");
             Properties.SetString ("RemoveTracksActionLabel", Catalog.GetString ("Remove From Play Queue"));
@@ -91,12 +93,10 @@ namespace Banshee.PlayQueue
             ServiceManager.PlayerEngine.ConnectEvent (OnPlayerEvent);
             ServiceManager.PlaybackController.TrackStarted += OnTrackStarted;
 
-            ServiceManager.SourceManager.AddSource (this);
-            
             // TODO change this Gtk.Action code so that the actions can be removed.  And so this
             // class doesn't depend on Gtk/ThickClient.
             actions = new PlayQueueActions (this);
-            
+
             Properties.SetString ("ActiveSourceUIResource", "ActiveSourceUI.xml");
             Properties.SetString ("GtkActionPath", "/PlayQueueContextMenu");
 
@@ -105,40 +105,51 @@ namespace Banshee.PlayQueue
             ServiceManager.SourceManager.MusicLibrary.TracksDeleted += HandleTracksDeleted;
             ServiceManager.SourceManager.VideoLibrary.TracksChanged += HandleTracksChanged;
             ServiceManager.SourceManager.VideoLibrary.TracksDeleted += HandleTracksDeleted;
-            
+
             populate_from = ServiceManager.SourceManager.Sources.FirstOrDefault (
-                source => source.Name == populate_from_name) as ITrackModelSource;
+                source => source.Name == populate_from_name) as DatabaseSource;
             if (populate_from != null) {
                 populate_from.Reload ();
             }
 
             TrackModel.Reloaded += HandleReloaded;
 
-            Offset = CurrentOffsetSchema.Get ();
+            int saved_offset = DatabaseConfigurationClient.Client.Get (CurrentOffsetSchema, CurrentOffsetSchema.Get ());
+            Offset = Math.Min (
+                saved_offset,
+                ServiceManager.DbConnection.Query<long> (@"
+                    SELECT MAX(ViewOrder) + 1
+                    FROM CorePlaylistEntries
+                    WHERE PlaylistID = ?", DbId));
+
+            ServiceManager.SourceManager.AddSource (this);
         }
 
         protected override void Initialize ()
         {
             base.Initialize ();
 
+            shuffler = new Shuffler (UniqueId);
             InstallPreferences ();
             header_widget = CreateHeaderWidget ();
             header_widget.ShowAll ();
 
             Properties.Set<Gtk.Widget> ("Nereid.SourceContents.HeaderWidget", header_widget);
         }
-        
+
         public HeaderWidget CreateHeaderWidget ()
         {
-            var header_widget = new HeaderWidget (populate_mode, populate_from_name);
-            header_widget.ModeChanged += delegate (object sender, ModeChangedEventArgs e) {
-                populate_mode = e.Value;
-                PopulateModeSchema.Set ((int) e.Value);
+            var header_widget = new HeaderWidget (shuffler, populate_shuffle_mode, populate_from_name);
+            header_widget.ModeChanged += delegate (object sender, EventArgs<RandomBy> e) {
+                populate_shuffle_mode = e.Value.Id;
+                PopulateModeSchema.Set (populate_shuffle_mode);
                 UpdatePlayQueue ();
                 OnUpdated ();
             };
 
-            header_widget.SourceChanged += delegate (object sender, SourceChangedEventArgs e) {
+            populate_shuffle_mode = header_widget.ShuffleModeId;
+
+            header_widget.SourceChanged += delegate (object sender, EventArgs<DatabaseSource> e) {
                 populate_from = e.Value;
                 if (populate_from == null) {
                     populate_from_name = String.Empty;
@@ -167,7 +178,7 @@ namespace Banshee.PlayQueue
         {
             EnqueueId (DatabaseTrackInfo.GetTrackIdForUri (uri), prepend, false);
         }
-        
+
         public void EnqueueTrack (TrackInfo track, bool prepend)
         {
             DatabaseTrackInfo db_track = track as DatabaseTrackInfo;
@@ -177,7 +188,7 @@ namespace Banshee.PlayQueue
                 EnqueueUri (track.Uri.AbsoluteUri, prepend);
             }
         }
-        
+
         private void EnqueueId (int trackId, bool prepend, bool generated)
         {
             if (trackId <= 0) {
@@ -226,14 +237,18 @@ namespace Banshee.PlayQueue
                 DbId, trackId, view_order, generated ? 1 : 0
             );
 
+            if (!generated) {
+                shuffler.RecordShuffleModification (trackId, ShuffleModificationType.Insertion);
+            }
+
             OnTracksAdded ();
             NotifyUser ();
         }
-        
+
         IDBusExportable IDBusExportable.Parent {
             get { return ServiceManager.SourceManager; }
         }
-        
+
         string IService.ServiceName {
             get { return "PlayQueue"; }
         }
@@ -279,6 +294,8 @@ namespace Banshee.PlayQueue
                     DbId, current_view_order
                 ));
 
+                WithTrackSelection (model, shuffler.RecordInsertions);
+
                 // Add the tracks to the end of the queue.
                 WithTrackSelection (model, AddTrackRange);
 
@@ -297,6 +314,7 @@ namespace Banshee.PlayQueue
                 // mark the first added track as current.
                 if (index != -1 && view_order == current_view_order) {
                     SetCurrentTrack (TrackModel[index] as DatabaseTrackInfo);
+                    SetAsPlaybackSourceUnlessPlaying ();
                 }
                 return true;
             }
@@ -306,12 +324,34 @@ namespace Banshee.PlayQueue
         private void SetAsPlaybackSourceUnlessPlaying ()
         {
             if (current_track != null && ServiceManager.PlaybackController.Source != this) {
-                PriorSource = ServiceManager.PlaybackController.Source;
-                ServiceManager.PlaybackController.NextSource = this;
+                bool set_source = !ServiceManager.PlayerEngine.IsPlaying ();
+                if (!set_source) {
+                    long view_order = ServiceManager.DbConnection.Query<long> (@"
+                        SELECT ViewOrder
+                        FROM CorePlaylistEntries
+                        WHERE PlaylistID = ? AND EntryID = ?",
+                        DbId, Convert.ToInt64 (current_track.CacheEntryId));
+                    long nongenerated = ServiceManager.DbConnection.Query<long> (@"
+                        SELECT COUNT(*)
+                        FROM CorePlaylistEntries
+                        WHERE PlaylistID = ? AND ViewOrder >= ? AND Generated = 0",
+                        DbId, view_order);
+                    set_source = nongenerated > 0;
+                }
+
+                if (set_source) {
+                    PriorSource = ServiceManager.PlaybackController.Source;
+                    ServiceManager.PlaybackController.NextSource = this;
+                }
             }
         }
 
         public void Clear ()
+        {
+            Clear (false);
+        }
+
+        private void Clear (bool disposing)
         {
             ServiceManager.DbConnection.Execute (@"
                 DELETE FROM CorePlaylistEntries
@@ -320,6 +360,10 @@ namespace Banshee.PlayQueue
             offset = 0;
             SetCurrentTrack (null);
 
+            if (disposing) {
+                return;
+            }
+
             if (this == ServiceManager.PlaybackController.Source && ServiceManager.PlayerEngine.IsPlaying ()) {
                 ServiceManager.PlayerEngine.Close();
             }
@@ -330,7 +374,7 @@ namespace Banshee.PlayQueue
         public void Dispose ()
         {
             int track_index = current_track == null ? Count : Math.Max (0, TrackModel.IndexOf (current_track));
-            CurrentTrackSchema.Set (track_index);
+            DatabaseConfigurationClient.Client.Set (CurrentTrackSchema, track_index);
 
             ServiceManager.PlayerEngine.DisconnectEvent (OnPlayerEvent);
             ServiceManager.PlaybackController.TrackStarted -= OnTrackStarted;
@@ -349,17 +393,17 @@ namespace Banshee.PlayQueue
             }
 
             if (!Populate && ClearOnQuitSchema.Get ()) {
-                Clear ();
+                Clear (true);
             }
         }
-        
+
         private void BindToDatabase ()
         {
             int result = ServiceManager.DbConnection.Query<int> (
                 "SELECT PlaylistID FROM CorePlaylists WHERE Special = 1 AND Name = ? LIMIT 1",
                 special_playlist_name
             );
-            
+
             if (result != 0) {
                 DbId = result;
             } else {
@@ -400,6 +444,7 @@ namespace Banshee.PlayQueue
 
         protected override void RemoveTrackRange (DatabaseTrackListModel model, RangeCollection.Range range)
         {
+            shuffler.RecordShuffleModifications (model, range, ShuffleModificationType.Discard);
             base.RemoveTrackRange (model, range);
 
             model.Selection.UnselectRange (range.Start, range.End);
@@ -412,7 +457,7 @@ namespace Banshee.PlayQueue
 
         private void HandleReloaded(object sender, EventArgs e)
         {
-            int track_index = CurrentTrackSchema.Get ();
+            int track_index = DatabaseConfigurationClient.Client.Get (CurrentTrackSchema, CurrentTrackSchema.Get ());
             if (track_index < Count) {
                 SetCurrentTrack (TrackModel[track_index] as DatabaseTrackInfo);
             }
@@ -455,6 +500,12 @@ namespace Banshee.PlayQueue
                         ServiceManager.PlaybackController.StopWhenFinished = true;
                     }
                 }
+                if (ServiceManager.PlaybackController.StopWhenFinished) {
+                    if (current_track != null && ServiceManager.PlayerEngine.CurrentTrack == current_track) {
+                        int index = TrackModel.IndexOf (current_track) + 1;
+                        SetCurrentTrack (index < Count ? TrackModel[index] as DatabaseTrackInfo : null);
+                    }
+                }
             } else if (args.Event == PlayerEvent.StartOfStream) {
                 if (TrackModel.IndexOf (ServiceManager.PlayerEngine.CurrentTrack) != -1) {
                     SetCurrentTrack (ServiceManager.PlayerEngine.CurrentTrack as DatabaseTrackInfo);
@@ -486,10 +537,10 @@ namespace Banshee.PlayQueue
 
         bool IBasicPlaybackController.First ()
         {
-            return ((IBasicPlaybackController)this).Next (false);
+            return ((IBasicPlaybackController)this).Next (false, true);
         }
-        
-        bool IBasicPlaybackController.Next (bool restart)
+
+        bool IBasicPlaybackController.Next (bool restart, bool userRequested)
         {
             if (current_track != null && ServiceManager.PlayerEngine.CurrentTrack == current_track) {
                 int index = TrackModel.IndexOf (current_track) + 1;
@@ -500,17 +551,24 @@ namespace Banshee.PlayQueue
                 ServiceManager.PlaybackController.Source = PriorSource;
                 if (was_playing) {
                     ServiceManager.PlaybackController.PriorTrack = prior_playback_track;
-                    ServiceManager.PlaybackController.Next (restart);
+                    ServiceManager.PlaybackController.Next (restart, userRequested);
                 } else {
+                    if (!userRequested) {
+                        ServiceManager.PlayerEngine.SetNextTrack ((TrackInfo)null);
+                    }
                     ServiceManager.PlayerEngine.Close ();
                 }
                 return true;
             }
 
-            ServiceManager.PlayerEngine.OpenPlay (current_track);
+            if (userRequested) {
+                ServiceManager.PlayerEngine.OpenPlay (current_track);
+            } else {
+                ServiceManager.PlayerEngine.SetNextTrack (current_track);
+            }
             return true;
         }
-        
+
         bool IBasicPlaybackController.Previous (bool restart)
         {
             if (current_track != null && ServiceManager.PlayerEngine.CurrentTrack == current_track) {
@@ -522,7 +580,7 @@ namespace Banshee.PlayQueue
             }
             return true;
         }
-        
+
         private void UpdatePlayQueue ()
         {
             // Find the ViewOrder of the current_track.
@@ -568,17 +626,11 @@ namespace Banshee.PlayQueue
                 // Add songs from the selected source, skip if all tracks need to be populated.
                 bool skip = tracks_to_add == upcoming_songs_number;
                 for (int i = 0; i < tracks_to_add; i++) {
-                    var track = populate_from.TrackModel.GetRandom (
-                        source_set_at, populate_mode, false, skip && i == 0) as DatabaseTrackInfo;
+
+                    var track = populate_from.DatabaseTrackModel.GetRandom (
+                        source_set_at, populate_shuffle_mode, false, skip && i == 0, shuffler) as DatabaseTrackInfo;
+
                     if (track != null) {
-                        track.LastPlayed = DateTime.Now;
-                        // track.Save() is quite slow, update LastPlayedStamp directly in the database.
-                        ServiceManager.DbConnection.Execute (@"
-                            UPDATE CoreTracks
-                            SET LastPlayedStamp = ?
-                            WHERE TrackID = ?",
-                            Hyena.DateTimeUtil.ToTimeT (track.LastPlayed), track.TrackId
-                        );
                         EnqueueId (track.TrackId, false, true);
                     }
                 }
@@ -612,7 +664,7 @@ namespace Banshee.PlayQueue
             protected set {
                 if (value != offset) {
                     offset = value;
-                    CurrentOffsetSchema.Set ((int) offset);
+                    DatabaseConfigurationClient.Client.Set (CurrentOffsetSchema, (int)offset);
                     Reload ();
                 }
             }
@@ -689,8 +741,10 @@ namespace Banshee.PlayQueue
             SetAsPlaybackSourceUnlessPlaying ();
         }
 
+        private bool enable_population = true;
         public bool Populate {
-            get { return populate_mode != PlaybackShuffleMode.Linear; }
+            get { return enable_population && populate_shuffle_mode != "off"; }
+            set { enable_population = value; }
         }
 
         private ITrackModelSource PriorSource {
@@ -709,6 +763,18 @@ namespace Banshee.PlayQueue
             }
         }
 
+        public override int EnabledCount {
+            get {
+                return current_track == null ? 0 : Count - TrackModel.IndexOf (current_track);
+            }
+        }
+
+        public override int FilteredCount {
+            get {
+                return EnabledCount;
+            }
+        }
+
         public override bool CanRename {
             get { return false; }
         }
@@ -779,23 +845,26 @@ namespace Banshee.PlayQueue
             "Clear the play queue when quitting"
         );
 
-        public static readonly SchemaEntry<int> CurrentTrackSchema = new SchemaEntry<int> (
+        // TODO: By 1.8 next two schemas can be removed. They are kept only to ease
+        // the migration from GConfConfigurationClient to DatabaseConfigurationClient.
+
+        private static readonly SchemaEntry<int> CurrentTrackSchema = new SchemaEntry<int> (
             "plugins.play_queue", "current_track",
             0,
             "Current Track",
             "Current track in the Play Queue"
         );
 
-        public static readonly SchemaEntry<int> CurrentOffsetSchema = new SchemaEntry<int> (
+        private static readonly SchemaEntry<int> CurrentOffsetSchema = new SchemaEntry<int> (
             "plugins.play_queue", "current_offset",
             0,
             "Current Offset",
             "Current offset of the Play Queue"
         );
 
-        public static readonly SchemaEntry<int> PopulateModeSchema = new SchemaEntry<int> (
-            "plugins.play_queue", "populate_mode",
-            (int) PlaybackShuffleMode.Linear,
+        public static readonly SchemaEntry<string> PopulateModeSchema = new SchemaEntry<string> (
+            "plugins.play_queue", "populate_shuffle_mode",
+            "off",
             "Play Queue population mode",
             "How (and if) the Play Queue should be randomly populated"
         );
